/*\
|*| --------------------------------------------------------------------------
|*|	S3 Vision964 chipset driver
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
|*|
|*|	This code was derived from the following source of information:
|*|	[1]	S3 Vision964 Data Book, (c) 1994 S3 Incorporated
\*/

/*\
|*|	Kernel includes
\*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <asm/io.h>

extern void request_region(unsigned int from, unsigned int num,
                           const char *name);
extern int  check_region  (unsigned int from, unsigned int num);
extern void release_region(unsigned int from, unsigned int num);

/*\
|*|     GGI includes
\*/

#undef DEBUG_IO
#define DRIVER_VERSION	"v0.00 ALPHA"

#include "graphdev.h"
#include "kgi-module.h"
#include "S3/Vision964.h"

#define IO_BASE   0x03B0
#define IO_REGION 0x03E0-IO_BASE

static unsigned short INS1r;
static unsigned short CRTI,CRTD;
static unsigned short FCTRLw;

static void
outbMISC(unsigned char value)
{
  outb(value, MISCw);			/* put value      */
  if (value & MISC_COLOR_IO)		/* adjust IO base */
    { CRTI=CRTIc; CRTD=CRTDc; INS1r=INS1rc; FCTRLw=FCTRLwc; }
   else
    { CRTI=CRTIm; CRTD=CRTDm; INS1r=INS1rm; FCTRLw=FCTRLwm; };
}

static inline void
outbCRT(unsigned char value, unsigned char crt_reg)
 {
   outb(crt_reg, CRTI); outb(value, CRTD);
   #ifdef DEBUG_IO
     TRACE(printk("CRT_%.2x <- %.2x\n", crt_reg, value));
   #endif
 }

static inline unsigned char
inbCRT(unsigned char crt_reg)
 { unsigned char foo;
   outb(crt_reg, CRTI); foo = inb(CRTD);
   #ifdef DEBUG_IO
     TRACE(printk("CRT_%.2x -> %.2x\n", crt_reg, foo));
   #endif
   return foo;
 }

static void
s3_unlock(void)
{
  outbCRT(CR38_MAGIC, 0x38);		/* unlock CR30-CR3F */
  outbCRT(CR39_MAGIC, 0x39);		/* unlock CR40-CRFF */
}

static void 
s3_lock(void)
{
  outbCRT(0, 0x38);			/* lock CR30-CR3F */
  outbCRT(0, 0x39);			/* lock CR40-CRFF */
}


static unsigned short dac_port[4] ={ 0x3C8, 0x3C9, 0x3C6, 0x3C7 };

void
outbDAC(unsigned char value, unsigned char dac_reg)
 {
   if(dac_reg & ~0xF)
     printk("%s: warning: DAC register select lines truncated: %x to %x.\n",\
            __FILE__, dac_reg, dac_reg &= 0xF);

   outb(0x55,CRTI); outb((inb(CRTD) & ~CR55_DAC_MASK) | (dac_reg >> 2), CRTD);
   outb(value, dac_port[dac_reg & 0x03]);

   #ifdef DEBUG_IO
     TRACE(printk("DAC_%.1x <- %.2x\n", dac_reg, value));
   #endif
 }

unsigned char
inbDAC(unsigned char dac_reg)
 {
   #ifdef DEBUG_IO
    unsigned char foo;
   #endif
  
   if(dac_reg & ~0x0F) 
     printk("%s: warning: DAC register select lines truncated: %x to %x.\n",\
	    __FILE__, dac_reg, dac_reg &= 0x0F);

   outb(0x55,CRTI); outb((inb(CRTD) & ~CR55_DAC_MASK) | (dac_reg >> 2), CRTD);

   #ifdef DEBUG_IO
     foo = inb(dac_port[dac_reg & 0x03]);
     TRACE(printk("DAC_%.1x -> %.2x\n",dac_reg, foo));
     return foo;
   #else
     return inb(dac_port[dac_reg & 0x03]);
   #endif
 }

void
insbDAC(unsigned short dac_reg, void *buf, int count)
{
  if(dac_reg & ~0x0F)
    printk("%s: warning; DAC register select lines truncated: %x to %x.\n",\
           __FILE__, dac_reg, dac_reg &= 0xF);

  outb(0x55, CRTI); outb((inb(CRTD) & ~CR55_DAC_MASK) | (dac_reg >> 2), CRTD);
  insb(dac_port[dac_reg & 0x03], buf, count);
}

void
outsbDAC(unsigned short dac_reg, void *buf, int count)
{
  if(dac_reg & ~0x0F) 
    printk("%s: warning: DAC register select lines truncated: %x to %x.\n",\
           __FILE__, dac_reg, dac_reg &= 0x0F);
  outb(0x55, CRTI); outb((inb(CRTD) & ~CR55_DAC_MASK) | (dac_reg >> 2), CRTD);
  outsb(dac_port[dac_reg & 0x03], buf, count);
}


void
setCLK(int value)
{
  if(value & ~CR42_DCLK_MASK)
    printk("%s: warning: clock select lines truncated: %x to %x.\n",\
           __FILE__, value, value &= CR42_DCLK_MASK);
  outbCRT((inbCRT(0x42) & ~CR42_DCLK_MASK) | value, 0x42);
}

int
getCLK(void)
{
  return inbCRT(0x5C) & CR5C_CLOCK_MASK;
}

#define NoSEQRegs	0x05
#define	NoCRTRegs	0x80
#define NoGRCRegs	0x09
#define NoATCRegs	0x15

struct statetable
{
   unsigned char MISC;
   unsigned char FCTRL;
   unsigned char SEQ[NoSEQRegs];
   unsigned char CRT[NoCRTRegs];
   unsigned char GRC[NoGRCRegs];
   unsigned char ATC[NoATCRegs];
};

static struct statetable initial_state;

static void
save_SEQ(unsigned char *buf)
{ int i;
  for(i = 0; i < NoSEQRegs; i++)
   {
     outb(i, SEQI); *(buf++)=inb(SEQD);
     #ifdef DEBUG_IO
       if (!(i & 0xF))
         printk("\n%.2x: ",i);
       printk("%.2X, ",*(buf-1));
     #endif
   }
  #ifdef DEBUG_IO
    printk("\n");
  #endif
}

static void
restore_SEQ(unsigned char *buf)
{ register unsigned char i;
  for(i=0; i < NoSEQRegs; i++)
   { outb(i,SEQI); outb(*(buf++), SEQD);}
}

static void
save_CRT(unsigned char *buf)
{ int i;
  for(i=0; i < NoCRTRegs; i++)
   { 
     outb(i, CRTI); *(buf++)=inb(CRTD);
     #ifdef DEBUG_IO
       if (!(i & 0xF))
         printk("\n%.2x: ",i);
       printk("%.2X, ",*(buf-1));
     #endif
   }
  #ifdef DEBUG_IO
    printk("\n");
  #endif
}

static void
restore_CRT(unsigned char buf[])
{ register unsigned char i;
  outbCRT(inbCRT(CRT_VSYNCEND) & ~CR11_LOCKTIMING, CRT_VSYNCEND);
  for(i=0x00; i < 0x19; i++) outbCRT(buf[i], i); 
  for(i=0x30; i < 0x3D; i++) outbCRT(buf[i], i);
  for(i=0x40; i < 0x6E; i++)
    switch(i) 
     {
       case 0x42:
	 outbCRT((inbCRT(i) & CR42_DCLK_MASK) | (buf[i] & ~CR42_DCLK_MASK), i); 
	 break;
       case 0x55:
	 outbCRT((inbCRT(i) & CR55_DAC_MASK) | (buf[i] & ~CR55_DAC_MASK), i);
         break;
       default: 
         outbCRT(buf[i], i);
     }
}

static void
save_GRC(unsigned char *buf)
{ int i;
  for(i=0; i < NoGRCRegs; i++) 
   {
     outb(i, GRCI); *(buf++)=inb(GRCD);
     #ifdef DEBUG_IO
       if (!(i & 0xF))
         printk("\n%.2x: ",i);
       printk("%.2X, ",*(buf-1));
     #endif
   }
  #ifdef DEBUG_IO
    printk("\n");
  #endif
}

static void
restore_GRC(unsigned char *buf)
{
  register unsigned char i;
  for(i = 0; i < NoGRCRegs; i++)
    {
      outb(i, GRCI); outb(*(buf++), GRCD);
      #ifdef DEBUG_IO
        if(!(i & 0xF))
          printk("\n%2.x: ",i);
        prink("%.2X, ",*(buf-1))
      #endif
    }
  #ifdef DEBUG_IO
    printk("\n");
  #endif
}

static void
save_ATC(unsigned char *buf)
{ int i;
  for(i=0; i < NoATCRegs; i++)
   { 
     inb(INS1r); outb(i, ATCI);
     *(buf++)=inb(ATCDr);  /* ^^ reset AFF,  ^^ setup index */
     #ifdef DEBUG_IO
       if (!(i & 0xF))
         printk("\n%.2x: ",i);
       printk("%.2X, ",*(buf-1));
     #endif
   }
  inb(INS1r); outb( ATCI_ENABLEDISPLAY, ATCI);
}

static void
restore_ATC(unsigned char *buf)
{
  register unsigned char i;
  /* inb(INS1r);	 initialize AFF */
  for(i = 0; i < NoATCRegs; i++)
   { 
     inb(INS1r); outb(i,ATCI); outb(*(buf++),ATCDw);
     #ifdef DEBUG_IO
       if(!(i & 0x0F))
         printk("\n%.2x: ",i);
       printk("%.2X, ",*(buf-1));
     #endif
   }
  inb(INS1r); outb(ATCI_ENABLEDISPLAY, ATCI);
}

static void
save_state(struct statetable *state)
 {
   s3_unlock();
   state->MISC  = inb(MISCr);	/* save miscellaneous output reg */
   #ifdef DEBUG_IO
     printk("MISC: %.2X\n",state->MISC);
   #endif
   state->FCTRL = inb(FCTRLr);	/* save feature control reg */
   save_SEQ(state->SEQ);	/* save sequencer regs */
   save_CRT(state->CRT);	/* save cathode ray tube controller regs */
   save_GRC(state->GRC);	/* save graphics controller regs */
   save_ATC(state->ATC);	/* save attribute controller regs */
 }

static void
restore_state(struct statetable *state)
 {
   outbMISC((inb(MISCr) & MISC_CLOCK_MASK) | state->MISC);
   outb(state->FCTRL, FCTRLw);
   restore_SEQ(state->SEQ);
   restore_CRT(state->CRT);
   restore_GRC(state->GRC);
   restore_ATC(state->ATC);
 }

#ifndef DEBUG_IO
static void
dump_state(struct statetable *state)
{ int i;
  printk("%.2x\n",state->MISC);
  printk("%.2x\n",state->FCTRL);
  for(i = 0; i < NoSEQRegs; i++)
   { if(!(i & 0xF)) printk("\n%.2x: ",i); printk("%.2x, ",state->SEQ[i]); }
  for(i = 0; i < NoCRTRegs; i++)
   { if(!(i & 0xf)) printk("\n%.2x: ",i); printk("%.2x, ",state->CRT[i]); }
  for(i = 0; i < NoGRCRegs; i++)
   { if(!(i & 0xF)) printk("\n%.2x: ",i); printk("%.2x, ",state->GRC[i]); }
  for(i = 0; i < NoATCRegs; i++)
   { if(!(i & 0xF)) printk("\n%.2x: ",i); printk("%.2x, ",state->ATC[i]); }
}
#endif

static struct GGI_ChipParams
chipinfo =
 {
   0,			/* installed RAM, entered during initialisation */
   2048, 2048,		/* register limits for maximum resolution */
   {16000000, 75000000},/* VCLK limits  */
   "S3 Incorporated",	/* manufacturer */
   "Vision 984",	/* model */

      GT_TEXT      	/* supported graphmodes */
    | GT_4BIT
    | GT_8BIT
    | GT_15BIT | GT_16BIT
    | GT_32BIT
 };

static struct statetable
default_state =
 {
     0x23, 0x00,					/* MISC, FCTRL */
   {
     0x03, 0x00, 0x0F, 0x00, 0x0E			/* SEQ */
   },
   {
     0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x10,	/* CRT */
     0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE3,		/* 0x1x */
     0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x2x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x09, 0x00, 0x20, 0x10, 0x00, 0x00, 0x00,		/* 0x3x */
     0x4A, 0xAD, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x4x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x00, 0x00,		/* 0x5x */
     0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x6x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x7x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
   },
   {
     0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F,	/* GRC */
     0xFF
   },
   {
     0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC */
     0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
     0xC0, 0x00, 0x0F, 0x00, 0x00
   }
 };

static int
detect_Vision964(void)
{
  #define COPYDEFAULT(X) default_state.CRT[(X)]=initial_state.CRT[(X)]
  register unsigned char foo;

  foo = COPYDEFAULT(0x30);
  if((foo & CR30_CHIPID) != CR30_Vision964ID)
    {
      printk("  Vision964 not detected (ID = %.2x).\n", foo & CR30_CHIPID);
      return FALSE;
    }
  TRACE(printk("  Vision964 rev %i detected.\n", foo & CR30_REVISION));


  foo = COPYDEFAULT(0x36);
   switch(foo & CR36_MEMSIZE_MASK)
    {
      case CR36_1MB: chipinfo.ramsize = 1;	break;
      case CR36_2MB: chipinfo.ramsize = 2;	break;
      case CR36_3MB: chipinfo.ramsize = 3;	break;
      case CR36_4MB: chipinfo.ramsize = 4;	break;
      case CR36_6MB: chipinfo.ramsize = 6;	break;
      case CR36_8MB: chipinfo.ramsize = 8;	break;
      default:  chipinfo.ramsize = 1;
                printk("%s: can't determine VRAM size, assume 1MB.\n",__FILE__);
    }
   #ifdef DEBUG
    switch(foo & CR36_RAMMASK)
     {
       case CR36_FASTPAGERAM: printk("  %i MB FP RAM, ",  chipinfo.ramsize); break;
       case CR36_EDO_RAM:     printk("  %i MB EDO RAM, ", chipinfo.ramsize); break;
       default:		      printk("  %i MB unknown RAM, ", chipinfo.ramsize);
     }
    switch(foo & CR36_BUSMASK)
     {
       case CR36_PCI      :   printk("PCI bus\n");        break;
       case CR36_VESALOCAL:   printk("Vesa Local bus\n"); break;
       default:               printk("unknown bus\n");
     }
   #endif
   chipinfo.ramsize *= 0x100000; /* convert MB to Byte */

  COPYDEFAULT(CRT_CONFIG2);
  default_state.CRT[0x3A] |= (initial_state.CRT[0x3A] & (CR3A_PCIMASK | CR3A_REFRESHMASK));
  default_state.CRT[0x40] |= (initial_state.CRT[0x40] & (CR40_BUSMASK));
  default_state.CRT[0x53] |= (initial_state.CRT[0x53] & (CR53_PAR_SID)); 
  default_state.CRT[0x58] |= (initial_state.CRT[0x58] & (CR58_SAM_256 | CR58_VL_T1_LATCH));
/*  default_state.CRT[0x5C] |= (initial_state.CRT[0x5C] & (CR5C_GOUT_MASK)); */
  COPYDEFAULT(CRT_MEMCONTROL3);
  COPYDEFAULT(CRT_MEMCONTROL4);
  COPYDEFAULT(CRT_MEMCONTROL5);
  default_state.CRT[0x65] |= (initial_state.CRT[0x65] & (CR65_SAM_MASK)); 
  default_state.CRT[0x66] |= (initial_state.CRT[0x66] & (CR66_PCI_DISCONNECT));
  COPYDEFAULT(CRT_CONFIG3);

  /* temporary, using 64 KB LAW at 000a0000 */
  default_state.CRT[0x59] = 0x00;
  default_state.CRT[0x5A] = 0x0a;
  default_state.CRT[0x58] |= CR58_LAW_ENABLE;

  return TRUE;
  #undef COPYDEFAULT
}

/*\
|*|	GGI interface
\*/

int
kgi_chipset_init(void)
{
  int detected; 

  TRACE(printk("S3 964 Chipset Driver "DRIVER_VERSION"\n"));

  outb(S3_DISABLE, S3_SUBSYSTEM);	/* chip wakeup		*/
  outb(S3_WAKEUP,  S3_SETUP);
  outb(S3_ENABLE,  S3_SUBSYSTEM);

  outbMISC(inb(MISCr));			/* adjust IO base	*/

  save_state(&initial_state);		/* unlock extended regs */
					/* and save state	*/

  if((detected = detect_Vision964()))
    {
      register unsigned char foo;
      
      /* select CR42 to drive the clock select lines. */
      foo = inb(MISCr);
      if((foo & MISC_CLOCK_MASK) != MISC_EXT_CLOCK)
        { 
          outbCRT((foo &  MISC_CLOCK_MASK) >> 2, 0x42);
          outbMISC((foo & ~MISC_CLOCK_MASK) | MISC_EXT_CLOCK);
        }

      if(! check_region(IO_BASE, IO_REGION))
        { 
          request_region(IO_BASE, IO_REGION, "Vision964");
 	  return TRUE;
        } 
       else   
        {
	  printk("%s: warning: can't get IO region.\n",__FILE__);
          return TRUE;
        }
    }
   else
    return FALSE;
}

void
kgi_chipset_done(void)
{
  release_region(IO_BASE, IO_REGION);
  TRACE(printk("S3 Chipset Driver removed\n"));
  restore_state(&initial_state);
  printk("%.2X (atci)\n", inbCRT(0x24));
  dump_state(&initial_state);
  s3_lock();
}

void
kgi_chipset_get_parameters (struct GGI_ChipParams *params)
{
  *params = chipinfo;
}

int
kgi_chipset_check_timing (struct GGI_Timing *TM)
{
  if(TM->flags & FL_INTERLACE)
   {
     TRACE(printk("%s: sorry, interlaced modes not supported (yet).\n",__FILE__));
     return FALSE; 
   }
  if(TM->clock == 0x7FFFFFFF)	/* first call ??? */
   {
     int Bpp = 1;

     if(TM->tr.graphtype == GT_24BIT)
      TM->tr.graphtype = GT_32BIT;

     if(!(TM->tr.graphtype & chipinfo.graphtypes))
      {
	TRACE(printk("%s: graphtype %i not supported.\n",__FILE__,\
		     TM->tr.graphtype));
	return FALSE;
      }

     switch(TM->tr.graphtype)
      {
	case GT_TEXT:
	  break;

	case GT_32BIT:	Bpp +=4;	/* (4+2+1+1)/2 = 4   Bpp */
	case GT_15BIT:
	case GT_16BIT:	Bpp +=2;	/*   (2+1+1)/2 = 2   Bpp */
	case GT_8BIT:	Bpp +=1;	/*     (1+1)/2 = 1   Bpp */
	case GT_4BIT:			/*         1/2 = 0.5 Bpp */

	  if (TM->tr.xvirtual <=  640) TM->tr.xvirtual =  640; else
	  if (TM->tr.xvirtual <=  800) TM->tr.xvirtual =  800; else
	  if (TM->tr.xvirtual <= 1024) TM->tr.xvirtual = 1024; else
	  if (TM->tr.xvirtual <= 1152) TM->tr.xvirtual = 1152; else
	  if (TM->tr.xvirtual <= 1280) TM->tr.xvirtual = 1280; else
	  if (TM->tr.xvirtual <= 1600) TM->tr.xvirtual = 1600; else
	  if (TM->tr.xvirtual <= 2048) TM->tr.xvirtual = 2048; else
	   {
	     TRACE(printk("%s: %i horizontal pixels are too many.\n",__FILE__,\
			   TM->tr.xvirtual));
	     return FALSE;
	   }

	  if( ((TM->tr.xvirtual * TM->tr.yvirtual * Bpp)/2) > chipinfo.ramsize)
	   {
	     TRACE(printk("%s: not enough VRAM to do virtual %ix%i at %i bpp",\
		   __FILE__, TM->tr.xvirtual, TM->tr.yvirtual, Bpp*4));
	     return FALSE;
	   }

	  TM->tr.xvisible &= 0x7FFFFFF8; TM->tr.yvisible &= 0x7FFFFFF8;

          return TRUE;

	default:	/* this should not happen ... */

	  INTERNALERROR
	  return FALSE;
      }
     return TRUE;
   }
  else /* TM->clock == 0x7FFFFFFF */
   {
     switch(TM->tr.graphtype)
      {
	case GT_TEXT:
	  break;

	case GT_4BIT:
	case GT_8BIT:
	case GT_15BIT:
	case GT_16BIT:
	case GT_32BIT:
	  if (TM->clock < chipinfo.clock.min)
	   {
	     TRACE(printk("%s: DCLK too low.\n",__FILE__));
	     return FALSE;
	   }
	  if (TM->clockdiv > 5)
	   {
	     TRACE(printk("%s: %i:1 multiplex mode not supported",__FILE__,\
		   1 << TM->clockdiv));
	     return FALSE;
	   }
	  if (TM->clock > chipinfo.clock.max)
	   {
	     TRACE(printk("%s: DCLK too high, set to %i kHz",__FILE__,
	     TM->clock/1000));
	     TM->clock = chipinfo.clock.max;
	   }
	  break;

	default:	/* this should not happen ... */
	  INTERNALERROR
	  return FALSE;
      }
     return TRUE;
   }
}

/*

void
test( int x1, int y1)
{
  while (inw(0x9AE8) & 0xE000);
  outw(0xa000,0xbee8);
  outw(0x0027,0xbae8);
  outl(0x00808080,0xa6e8);
  outw(x1 & 0x0fff, 0x86e8);
  outw(y1 & 0x0fff, 0x82e8);
  outw(0x12FF,0x9ae8);
  while(inw(0x9ae8) & 0xf000);
  outw(0x0717,0x9ee8);
}
*/

void
test(void)
{
  int x,y;

  for(x = 0; x < 32; x++)
    for(y = 0; y < 32; y++)
     ((char *)0xa0000)[x+y*800] = x+y;
}
int
kgi_chipset_set_timing (struct GGI_Timing *TM)
{
  struct statetable setup = default_state;
  struct GGI_DacParams dac;
  register int foo;
  int DCLKdiv;
  int EC00 = EC00_IRQ_MASK;
  int EC01 = EC01_ENH_ENB | EC01_LAW_ENB;

  /*\
  |*|	Setup bits per pixel and logical screen width
  \*/

  switch(TM->tr.graphtype)
   {
     case GT_4BIT:
       setup.CRT[0x50] |= CR50_4BPP;
       EC01 |= EC01_4BPP;
       foo = TM->tr.xvirtual >> 4;
       break;
     case GT_8BIT:
       setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
       setup.CRT[0x50] |= CR50_8BPP;
       foo = TM->tr.xvirtual >> 3;
       break;
     case GT_15BIT:
     case GT_16BIT:
       setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
       setup.CRT[0x50] |= CR50_16BPP;
       setup.CRT[0x67] |= CR67_NO_INVERT_VCLK;
       foo = TM->tr.xvirtual >> 2;
       break;
     case GT_32BIT:
       setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
       setup.CRT[0x50] |= CR50_32BPP;
       foo = TM->tr.xvirtual >> 1;
       break;

     default:
       INTERNALERROR
       return FALSE;       
   }
   setup.CRT[CRT_LOGICALWIDTH]  |= foo & 0xFF;
   setup.CRT[0x51] |= (foo & 0x100) ? CR51_LOGICALWIDTH_B8 : 0;
   setup.CRT[0x51] |= (foo & 0x200) ? CR51_LOGICALWIDTH_B9 : 0;

   if(TM->tr.graphtype == GT_TEXT)
     {
       DCLKdiv = 8; /* 7,8,9 */ 
	/* setup SEQ according to DCLK div !!! */
     }
    else /* TM->tr.graphtype == GT_TEXT */
     {
       DCLKdiv = 8;

       setup.SEQ[0x01] |= SR01_8DOTCLK;
       setup.ATC[0x10] |= AR10_GRAPHICS;
       setup.GRC[0x06] |= GR6_GRAPHMODE;

       setup.CRT[0x09] |= TM->ymagnify & CR09_CHARHEIGHT_MASK;
       setup.CRT[0x09] |= (TM->flags & FL_DOUBLESCAN) ? CR09_DOUBLESCAN : 0;

       switch(TM->tr.xvirtual)
        { 
          case  640: setup.CRT[0x50] |= CR50_640; break;
          case  800: setup.CRT[0x50] |= CR50_800; break;
          case 1024: setup.CRT[0x50] |= CR50_1024; break;
          case 1152: setup.CRT[0x50] |= CR50_1152; break;
          case 1280: setup.CRT[0x50] |= CR50_1280; break;
          case 1600: setup.CRT[0x50] |= CR50_1600; break;
          case 2048: setup.CRT[0x50] |= CR50_2048;
		     setup.CRT[0x31] |= CR31_2PAGE_SCRN; 
		     break;
          default: INTERNALERROR return FALSE;
        }

       /*\
       |*|  ???  I am not sure whether the following is right. May be these bits
       |*|  ???  have to be set according to the ammount of VRAM installed ?
       |*|  ???  If anyone has such a hardware, please contact me.
       \*/

       kgi_ramdac_get_parameters(&dac);
       if(dac.dactype & DAC_VRAM)
         {
           setup.CRT[0x66] |= CR66_PIXELBUS_OFF;
           switch(dac.dactype & DAC_PB_MASK)
            { /* some are NOT_TESTED yet. If you have such a configuration
	       * please report your experience.  May be CR65 needs to be updated
	       * too ??? (FIXME) */
	      case DAC_16BIT_PB:
	 	setup.CRT[0x66] |= CR66_32S_SID;
	 	NOT_TESTED;
	 	break;
	      case DAC_32BIT_PB:
	 	setup.CRT[0x66] |= CR66_32P_SID;
/* ???	 	setup.CRT[0x53] |= CR53_PAR_SID; */
/* ???	 	setup.CRT[0x51] |= CR51_DISABLE_SPLIT;*/
	 	break;
	      case DAC_64BIT_PB:
	 	setup.CRT[0x66] |= CR66_64S_SID;
	 	NOT_TESTED;
		break; 
	      case DAC_128BIT_PB:
	 	setup.CRT[0x66] |= CR66_128S_SID;
	 	NOT_TESTED;
	 	break; 
	      default:
	 	printk("%s: Don't know how to handle DAC type %.8x.\n",\
	 		__FILE__,dac.dactype);
	 	return FALSE;
            }
         }
        else /* dac.dactype & DAC_VRAM */ 
         { /* strange: a 964 with a RAMDAC that doesn't support VRAM SCLK. */
           printk(\
          "%s: The selected RAMDAC doesn't support VRAM shift clock but this\n"\
          "%s: driver requires the DAC to use VRAM mode in graphics modes.\n"\
          "%s: This may be beause you selected an inappropriate RAMDAC for\n"\
          "%s: your card. If not, please report to linux-ggi@eskimo.com.\n",\
           __FILE__,__FILE__,__FILE__,__FILE__);
          return FALSE;
         }
     }

  /*\
  |*|	setup horizontal timing
  \*/
 
   foo = (TM->xwidth / DCLKdiv) - 1;		/* horizontal display end */
     setup.CRT[0x01] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HDISPLAYEND_B8 : 0;

   foo = (TM->xsyncstart / DCLKdiv) - 1;	/* horizontal blank start */
     setup.CRT[0x02] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HBLANKSTART_B8 : 0;

   foo = (TM->xsyncstart / DCLKdiv);		/* horizontal sync start  */
     setup.CRT[0x04] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HSYNCSTART_B8 : 0;

   foo = (TM->xsyncend / DCLKdiv);		/* horizontal sync end    */
     setup.CRT[0x05] |= foo & CR05_HSE_MASK;
     setup.CRT[0x5D] |= ((foo- (TM->xsyncstart/DCLKdiv)) > 0x20)
				 ? CR5D_HSYNCEND_B5 : 0;

   foo = (TM->xsyncend / DCLKdiv); 		/* horizontal blank end   */
     setup.CRT[0x03] |= foo & CR03_HBE_MASK;
     setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;
     setup.CRT[0x5D] |= ((foo - (TM->xsyncstart/DCLKdiv)) > 0x40)
				 ? CR5D_HBLANKEND_B6 : 0;

   foo = (TM->xend / DCLKdiv) - 5;		/* horizontal total       */
     setup.CRT[0x00] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HTOTAL_B8 : 0;

  /*\
  |*|	setup vertical timing
  \*/
 
   foo = TM->ywidth - 1;			/* vertical display end   */
     setup.CRT[0x12] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VDISPLAYEND_B10 : 0;

   foo = TM->ysyncstart;			/* vertical blank start   */
     setup.CRT[0x15] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8  : 0;
     setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VBLANKSTART_B10 : 0;
    
   foo = TM->ysyncstart;			/* vertical sync start    */
     setup.CRT[0x10] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VSYNCSTART_B10 : 0;

   foo = TM->ysyncend;				/* vertical sync end      */
     setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;
 
   foo = TM->ysyncend + 1;			/* end vertical blank     */
     setup.CRT[0x16] |= foo & 0xFF;
  
   foo = TM->yend - 2;				/* vertical total         */
     setup.CRT[0x06] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VTOTAL_B10 : 0;
 
   setup.CRT[CRT_INTERLACESTART]|= TM->yend >> 1; /* ??? may be wrong. (FIXME)*/
   if(TM->flags & FL_INTERLACE)
    {
       printk("%s:%i: no interlaced modes (yet).\n",__FILE__,__LINE__);
       setup.CRT[0x42] |= CR42_INTERLACED;
       return FALSE; /* temporary */
    }

  setup.MISC |= TM->xsyncpol ? MISC_NEG_HSYNC : 0;
  setup.MISC |= TM->ysyncpol ? MISC_NEG_VSYNC : 0;

  setup.CRT[0x66] |= TM->clockdiv & CR66_VCLK_DIV_MASK;

  /*\
  |*|	Setup the FIFO control parameters.
  \*/

  foo = (TM->xend + TM->xsyncstart)/2/DCLKdiv;
  setup.CRT[0x3B] |= foo & 0xFF;
  setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_FIFOSTART_B8 : 0;

  /*\
  |*|	As we got here it seems to be ok, thus do the programming now.
  \*/

/*  setup.CRT[0x3C] = 0x3D; */

  outbCRT(inb(CRT_EXTSYNC1) | CR56_SYNC_OFF, CRT_EXTSYNC1);
  kgi_monitor_set_timing(TM);
  kgi_ramdac_set_timing(TM);
  kgi_clock_set_timing(TM);
  restore_state(&setup);

  outb(EC00 | EC00_RESET, ECR_MODECONTROL);
  outb(EC00, ECR_MODECONTROL);
  outb(EC01, ECR_FCTRL);
  
  #ifdef DEBUG_IO
    dump_state(&setup);
  #endif

  outw(0x1000,0xbee8);	/* temporary */
  outw(0x2000,0xbee8);
  outw(0x3fff,0xbee8);
  outw(0x4fff,0xbee8);

  return TRUE;
}

int
kgi_chipset_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  return -1;
}
