/*\
|*| ---------------------------------------------------------------------------
|*|	ICD 2061A Clock setter
|*| ---------------------------------------------------------------------------
|*|
|*|	1st version by Andreas Beck   [becka@hp.rz.uni-duesseldorf.de]
|*|	   modified by Steffen Seeger [seeger@physik.tu-chemnitz.de]
|*|
|*|	Copyright (C) 1995 Andreas Beck, Steffen Seeger
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*| ---------------------------------------------------------------------------
|*|     This code was derived from the following sources of information:
|*|	[1]	IC Designs data sheet for the ICD2061A V2.5, 12-Sep-93
|*|     [2]     XFree86 3.0 package, files ICD2061.h,ICD2061Aalt.c
|*|             (partial Copyright (c) 1992 by Number Nine Corp.)
\*/

#include <linux/kernel.h>

#include "graphdev.h"
#include "kgi.h"

/*\
|*| Clock driver dependend #define's
\*/

#define CLK_VERSION "0.11"

#define	REFERENCE_CLK	14318180 /* Hz, must be within 1..60 MHz */ 
static unsigned fref = REFERENCE_CLK * 2;		/* Hz */

#define MIN_FVCO  50000000	/* Hz */
#define MAX_FVCO 120000000	/* Hz */

#define	MIN_FREQ clock.clock.range[0].min
#define	MAX_FREQ clock.clock.range[0].max 

static struct kgi_clock clock =
{
	"IC Designs",		/* manufacturer		*/ 
	"ICD 2061A",		/* model		*/
	GM_ALL,			/* supported modes 	*/
	0,			/* special capabilities	*/
	{ 0x7FFF, 0x7FFF },	/* maximum resolution	*/

	CLK_TYPE_PROG,		/* clock type		*/
	{{   400000, 119000000,	/* clock ranges		*/
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
                  0,         0}}
};


#define LAST_INDEX 14
static int range[LAST_INDEX+1]=		/* VCO clock ranges in Hz */
 {  50000000, 51000000, 53200000, 58500000,
    60700000, 64400000, 66800000, 73500000,
    75600000, 80900000, 83200000, 91500000,
   100000000, 120000000, 120000000 };


/* This calculates the DCLK frequency for a given ratio and VCO clock.      */

#define F(P,Q,VCO)   VCO*(P+3)/(Q+2)


/* These give clock commands, that ain't used by the icd2061.               */

#define INVALIDCLK   -1
#define	INVALIDP    0xFFFF
#define INVALIDQ    0xFFFF


static int clk_last_freq = 0;		/* some kind of 'cache' */
static int clk_last_cmd  = INVALIDCLK;

/*\
|*|	A icd2061a connected in "standard" fashion is programmed by
|*|	transferring the command word over the two clock select lines.
|*|	The chipset driver has to care about how to assert these lines.
\*/

#define ICD_BOTH 0x03	/* data and clock	      */
#define ICD_DATA 0x02	/* icd2061a data pin 	      */
#define ICD_CLK  0x01	/* icd2061a clock pin	      */
#define	ICD_NONE 0x00	/* neither DATA nor CLK       */

static int
setDCLK(int clock_cmd, int clk)
{
  if(clock_cmd != INVALIDCLK)
    {
	/*\
	|*|	program icd2061 command word into the clock chip
	|*|	a command word is as follows:
	|*|
	|*|	 0 -  6		q
	|*|	 7 -  9		div
	|*|	10 - 16		p
	|*|	17 - 20		clock index selector
	|*|	21 - 23		register number
	|*|
	|*|	clock rate is 2*REFERENCE_CLK*(p+3) / ((q+2)* 2^div)
	\*/

	register unsigned i;
	unsigned m;
      
	clock_cmd |= (clk << 21);

	setCLK(ICD_DATA);		/* do setup as described in [1]       */
	setCLK(ICD_NONE);
	setCLK(ICD_DATA);

	for (i=0; i<6; i++) { setCLK(ICD_BOTH); setCLK(ICD_DATA); };
	for (i=0; i<2; i++) { setCLK(ICD_NONE); setCLK(ICD_CLK ); };

	for (i=0, m=1; i<24; i++, m+=m)
	  if (clock_cmd & m)
            {
		setCLK(ICD_CLK ); setCLK(ICD_NONE);
		setCLK(ICD_DATA); setCLK(ICD_BOTH);
            }
	   else
            {
		setCLK(ICD_BOTH); setCLK(ICD_DATA);
		setCLK(ICD_NONE); setCLK(ICD_CLK );
            }

	setCLK(ICD_BOTH); setCLK(ICD_DATA);
	setCLK(ICD_BOTH); setCLK(ICD_DATA);

	/* usleep(3000);	 wait for 3 ms should be enough */
	setCLK(clk);	/* now select clock generator     */
	return EOK;
    }
   else
    return ECLOCK;
}


static unsigned int
bestratio(unsigned *p, unsigned *q, int fvco)
{
  int mindiff = 0x7FFFFFFF;
  int diff;

  unsigned n;

  *p = INVALIDP;			/* make p and q invalid               */
  *q = INVALIDQ;

  for(n = 15; n < 72; n++)		/* n = q must be in [15,71]           */
   {
     int M = MAX_FVCO*n / fref - 3;
     int m = MIN_FVCO*n / fref - 3;

     if (m <   0) m =   0;		/* p = m must be in [0,127]           */
     if (M > 128) M = 128;
     
     for(; m < M; m++)
      {
        diff = fvco - F(m,n,fref);

        if ((diff >= 0) && (mindiff > diff))
	 {
           mindiff = diff;
           *p = m;
           *q = n;
         }
      }
   }
  return mindiff;
}

static int 
calcfreq(int *freq)
{
  unsigned m;
  int fvco;
  struct {
     unsigned p,q,m,i;
     double   delta_vco;
     unsigned delta_f;
   } best = { INVALIDP, INVALIDQ, 0,0, 1e10, 0xFFFFFFFF };
  double delta_vco;
  unsigned delta_f = 0xFFFFFFFF;

  if(*freq == clk_last_freq)
    return clk_last_cmd;

  fvco = *freq;

  for(m=0; m < 8; m++)
   {
     int i,p,q;

     for(i = LAST_INDEX; (i > -1) && (fvco < range[i]); i--);

     if (i < 0)
       { fvco += fvco; continue; }

     if (i == LAST_INDEX)

       break;

     delta_vco = abs((fvco - (range[i]+range[i+1])/2.0) / (double)fvco);
     delta_f   = bestratio(&p, &q, fvco);

     if((delta_vco < best.delta_vco) || (delta_f < best.delta_f))
       {
         best.delta_f   = delta_f;
         best.delta_vco = delta_vco;
         best.p = p;
         best.q = q;
         best.m = m;
         best.i = i;
       }
   }

  if(best.p == INVALIDP)
    {
      TRACE(printk("%s: invalid DCLK request. (%i Hz)\n",__FILE__,*freq));
      return INVALIDCLK;
    }

  TRACE(printk("%s: request for DCLK = %i Hz,",__FILE__, *freq));

  fvco = fref >> best.m;
  clk_last_freq = *freq = F(best.p, best.q, fvco);

  TRACE(printk(" did DCLK = %i Hz.\n", *freq));
 
  return  clk_last_cmd = (best.i << 17) |	/* clock index	*/
			 (best.p << 10) |	/* p value	*/
                  	 (best.m <<  7) |	/* divider	*/
                   	 (best.q      );	/* q value	*/
}

static int clk_save;
static int clk = 2;

/*\
|*|     GGI clock driver interface
\*/

int
kgi_clock_init(void)
{
  TRACE(printk("%s %s clockchip driver v"CLK_VERSION".\n", \
		clock.manufact, clock.model));
  if((clk_save = getCLK()) > 1)
   {
     printk(
     "%s: warning: initial clock selector is %x (should be 0 or 1).\n"\
     "%s: Please reconfigure your system to startup with 80x25 mode.\n",\
     __FILE__, clk_save, __FILE__);
     return FALSE;
   }
  return TRUE;
}

void
kgi_clock_done(void)
{
  setCLK(clk_save);
  TRACE(printk("ICD 2061a clock driver removed.\n"));
}

struct kgi_clock
*kgi_clock_get_info(void)
{
  return &clock;
}

int
kgi_clock_check_mode(struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int newfreq = mode->dclk;

  if (mode->dclk < MIN_FREQ)
   {
     TRACE(printk("%s: requested DCLK too low (%i Hz)\n",__FILE__, mode->dclk));
     return ETM_DCLK_LOW;
   }

  switch(cmd)
    {
      case TC_PROPOSE_TIMING:

   	if (mode->dclk > MAX_FREQ)	/* meet limits */
	  mode->dclk = MAX_FREQ;

	return (calcfreq(&(mode->dclk)) != INVALIDCLK) ? EOK : ECLOCK;

      case TC_CHECK_TIMING:

	if (mode->dclk > MAX_FREQ)
	  return ETM_DCLK_HIGH;

	if (   (calcfreq(&newfreq) != INVALIDCLK)
	    && ((mode->dclk - newfreq) < (mode->dclk*3/100)))
	  {
	    mode->dclk = newfreq;
	    return EOK;
	  };

	return ECLOCK;

      default:
	INTERNALERROR;
	return ECLOCK;
    }
}

int
kgi_clock_set_mode(struct kgi_mode *mode)
{
  int clock_cmd = calcfreq(&mode->dclk);

  setDCLK(clock_cmd, clk);
  return EOK;
}

int
kgi_clock_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	       unsigned long arg)
{
  return ECLOCK;
}
