/*\
|*| --------------------------------------------------------------------------
|*|     GGI device file definitions
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Andreas Beck, [becka@hp.rz.uni-duesseldorf.de]
|*|			  and Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#ifndef _graphdev_h
#define _graphdev_h

#include <linux/errno.h>
#include <linux/fs.h>

#include "ggi.h"
#include "err.h"

/*\
|*|	Definitions for device-registration
\*/

#define GRAPHDEV_MAJOR 8
#define GRAPHDEV_NAME  "graphics"

/*\
|*|	The ioctl's
\*/

#define IOCTL_MAGIC_MASK	0xff000000
#define IOCTL_MAGIC		0xab000000

#define IOCTL_SUBSYSTEM_MASK	0x00ff0000		

#define IOCTL_DRIVER		0x00010000      /* kernel driver  */
#define IOCTL_MONITOR		0x00020000      /* monitor driver */
#define IOCTL_RAMDAC		0x00030000      /* ramdac driver  */
#define IOCTL_CLOCK		0x00040000      /* clock driver   */
#define IOCTL_CHIPSET		0x00050000      /* chipset driver */
#define IOCTL_GRAPHICS		0x00060000      /* graphic driver */

/* ioctls for the graphics device : */

#define IDR	(IOCTL_DRIVER|IOCTL_MAGIC)	/* kernel driver  */

#define IO_DRIVER_SETMODE	(IDR|0x0001)		/* Required */
#define IO_DRIVER_GETMODE	(IDR|0x0002)		/* Required */
#define IO_DRIVER_GETSTATESIZE	(IDR|0x0003)		/* Required */
#define IO_DRIVER_SAVESTATE	(IDR|0x0004)		/* Required */
#define IO_DRIVER_RESTORESTATE	(IDR|0x0005)		/* Required */
#define IO_DRIVER_BGMODE	(IDR|0x0006)		/* Required */
#define IO_DRIVER_FGMODE	(IDR|0x0007)		/* Required */
#define IO_DRIVER_KEEPMODE	(IDR|0x0008)		/* suser only */
#define IO_DRIVER_SETDRAWINFO	(IDR|0x0009)		/* Required must be executed before 
							   context-sensitive operations */
#define IO_DRIVER_SETDEFFONT	(IDR|0x000a)		/* suser only */
#define IO_DRIVER_GETDEFFONT	(IDR|0x000b)		/* Required   */

#define IO_DRIVER_GETVT		(IDR|0x000c)		/* Required   */
#define IO_DRIVER_TEST		(IDR|0x000d)		/* This will only return 0 */


#define ICH	(IOCTL_CHIPSET|IOCTL_MAGIC)		/* kernel driver  */

#define IO_CHIP_SETVISWIN	(ICH|0x0001)		/* Required */
#define IO_CHIP_GETVISWIN	(ICH|0x0002)		/* Required */
#define	IO_CHIP_SETSPLITLINE	(ICH|0x0003)		/* optional */
#define	IO_CHIP_GETSPLITLINE	(ICH|0x0004)		/* optional */
#define IO_CHIP_SETPAGE		(ICH|0x0005)		/* Required */
#define IO_CHIP_GETPAGE		(ICH|0x0006)		/* Required */
#define IO_CHIP_WAITHRETRACE	(ICH|0x0007)		/* Optional */
#define IO_CHIP_WAITVRETRACE	(ICH|0x0008)		/* Optional */

#define IGR	(IOCTL_GRAPHICS|IOCTL_MAGIC)		/* kernel driver  */

#define IO_GRAPH_DRAWPIXEL	(IGR|0x0200)		/* Required */
#define IO_GRAPH_GETPIXEL	(IGR|0x0201)		/* Required */
#define IO_GRAPH_PUTPIXEL	(IGR|0x0202)		/* Required */
#define IO_GRAPH_SET16COL	(IGR|0x0203)		/* Required for 16 col modes */

							/* Optional from now on */
#define IO_GRAPH_DRAWHLINE	(IGR|0x0208)
#define IO_GRAPH_GETHLINE	(IGR|0x0209)
#define IO_GRAPH_PUTHLINE	(IGR|0x020a)

#define IO_GRAPH_DRAWVLINE	(IGR|0x0210)
#define IO_GRAPH_GETVLINE	(IGR|0x0211)
#define IO_GRAPH_PUTVLINE	(IGR|0x0212)

#define IO_GRAPH_DRAWBOX	(IGR|0x0220)
#define IO_GRAPH_GETBOX		(IGR|0x0221)
#define IO_GRAPH_PUTBOX		(IGR|0x0222)		/* Display Sprite */

#define IO_GRAPH_FILLSCREEN	(IGR|0x0228)

#define IO_GRAPH_DRAWLINE	(IGR|0x0240)

#define IO_GRAPH_DRAWCIRCLE	(IGR|0x0244)

/* Ioctls for the ramdac */
#define IDAC	(IOCTL_RAMDAC|IOCTL_MAGIC)		/* kernel driver  */

#define IO_RAMDAC_SETPAL	(IDAC|0x0001)		/* Required */
#define IO_RAMDAC_GETPAL	(IDAC|0x0002)		/* Required */
#define IO_RAMDAC_SETGAMMA	(IDAC|0x0003)		/* Optional */
#define IO_RAMDAC_GETGAMMA	(IDAC|0x0004)		/* Optional */
#define IO_RAMDAC_SETHWCDATA	(IDAC|0x0008)		/* Optional */
#define IO_RAMDAC_SETHWCPOS	(IDAC|0x0009)

/* Ioctls for the monitor */
#define IMON	(IOCTL_MONITOR|IOCTL_MAGIC)		/* kernel driver  */

#define	IO_MONITOR_GETINFO		(IMON|0x0000)
#define IO_MONITOR_SETINFO		(IMON|0x0001)		/* suser only */
#define IO_MONITOR_SETPREFMODE	(IMON|0x0010)		/* suser only */
#define IO_MONITOR_GETPREFMODE	(IMON|0x0011)
#define IO_MONITOR_CLEARPREF	(IMON|0x0012)		/* suser only */
#define IO_MONITOR_POWERSAVE	(IMON|0x0020)

/* mmap-constants for the offset (which is actually ignored) */

#define MMAP_LINEAR	0	/* Linear frame buffer */
#define MMAP_FRAME	1	/* banked frame buffer - optional ... for old apps */

/* MACROS : */

#define COPYPAR(size) \
	do{ int i; \
	    if ((i=verify_area(VERIFY_READ , (void *) arg, size))) return i; \
	    else memcpy_fromfs(parameters, (void *) arg, size); }while(0)

#endif	/* #ifdef _graphdev_h */
